/*! 
 * \file fwdpp/sugar/sugar.hpp
 * \brief Lazy header for fwdpp/sugar features
 */
#ifndef __FWDPP_SUGAR_SUGAR_HPP__
#define __FWDPP_SUGAR_SUGAR_HPP__

#include <fwdpp/sugar/popgenmut.hpp>
#include <fwdpp/sugar/add_mutation.hpp>

#endif
